from datetime import datetime

import transaction

from hattie.util import legistar_id_guid, make_true_date
from hattie.util import convert_range_to_datetime
from hattie.database import Meeting, Item, Attachment
from hattie.database import Action, ItemAction, ActionVote
from hattie.managers.base import BaseManager


class ItemManager(BaseManager):
    def query(self):
        return self.session.query(Item)

    def add(self, itemdata):
        with transaction.manager:
            i = Item()
            for key in itemdata:
                if key == 'attachments':
                    continue
                value = itemdata[key]
                setattr(i, key, value)
            self.session.add(i)
            if itemdata['attachments'] is not None:
                for name, link in itemdata['attachments']:
                    id, guid = legistar_id_guid(link)
                    if self.session.query(Attachment).get(id) is not None:
                        raise RuntimeError("Duplicate attachment %d" % id)
                    a = Attachment()
                    a.id = id
                    a.guid = guid
                    a.name = name
                    a.link = link
                    a.item_id = i.id
                    self.session.add(a)
        return self.session.merge(i)


class ActionManager(BaseManager):
    def query(self):
        return self.session.query(Action)

    def add(self, item_id, actiondata):
        with transaction.manager:
            a = Action()
            actiondata['filetype'] = actiondata['ftype']
            main_keys = ['id', 'guid']
            rollcall = actiondata['roll_call']
            if not rollcall:
                main_keys += ['action', 'action_text', 'result', 'filetype']
            for key in main_keys:
                setattr(a, key, actiondata[key])
            if not rollcall:
                for key in ['agenda_note', 'minutes_note']:
                    value = actiondata[key].strip()
                    if not value:
                        value = None
                    setattr(a, key, value)
                for key in ['mover', 'seconder']:
                    name, link = actiondata[key]
                    if not name:
                        continue
                    id, guid = legistar_id_guid(link)
                    attribute = '%s_id' % key
                    setattr(a, attribute, id)
                file_id, link = actiondata['file_id']
                a.file_id = file_id
            else:
                a.action = 'Roll Call'
            self.session.add(a)
            # flush here so the action can be referred by
            # foreign keys in 'item_action' and 'action_vote'
            self.session.flush()
            # make item_action object
            item_action = ItemAction(item_id, a.id)
            self.session.add(item_action)
            # handle votes
            for name, link, vote in actiondata['votes']:
                person_id, ignore = legistar_id_guid(link)
                avote = ActionVote(a.id, person_id, vote)
                self.session.add(avote)
        return self.session.merge(a)


class MeetingManager(BaseManager):
    def query(self):
        return self.session.query(Meeting)

    def _add(self, mdata):
        pass

    def add(self, mdata):
        pass

    def _range_filter(self, query, start, end):
        query = query.filter(Meeting.date >= start)
        query = query.filter(Meeting.date <= end)
        return query

    def get_meeting_list(self):
        query = self.session.query(Meeting).order_by(Meeting.date)
        return query.all()

    def get_ranged_meetings(self, start, end, timestamps=False):
        if timestamps:
            start, end = convert_range_to_datetime(start, end)
        q = self.session.query(Meeting)
        q = self._range_filter(q, start, end)
        return q.all()

    def _add_meeting_from_rss(self, entry):
        with transaction.manager:
            m = Meeting()
            m.title = entry.title
            m.link = entry.link
            m.rss = entry
            m.id, m.guid = legistar_id_guid(entry.link)
            m.updated = datetime.now()
            self.session.add(m)
        return self.session.merge(m)

    def _add_collected_meeting(self, meeting, collected):
        with transaction.manager:
            info = collected['info']
            for key in ['id', 'guid', 'time', 'link', 'dept_id',
                        'agenda_status', 'minutes_status']:
                value = info[key]
                setattr(meeting, key, value)
            meeting.date = make_true_date(info['date'])
            meeting.updated = datetime.now()
            m = self.session.merge(meeting)
        return m



import re

from ..util import onclick_link
from ..util import make_true_date
from ..util import legistar_id_guid


from .base import BaseCollector

ITEM_DATA_IDENTIFIERS = dict(file_id='_lblFile2',
                             name='_lblName2',
                             filetype='_lblType2',
                             status='_lblStatus2',
                             introduced='_lblIntroduced2',
                             on_agenda='_lblOnAgenda2',
                             attachments='_lblAttachments2',
                             title='_lblTitle2',
                             passed='_lblPassed2',
                             action_details='_hypDetails')


class ItemCollector(BaseCollector):
    def _get_item(self, page):
        markers = ITEM_DATA_IDENTIFIERS
        item_keys = list(markers.keys())
        item = {}.fromkeys(item_keys)
        item['action_details'] = []
        for key in item_keys:
            exp = re.compile('.+%s$' % markers[key])
            tags = page.find_all('span', id=exp)
            if not tags:
                tags = page.find_all('a', id=exp)
            if not tags and key == 'action_details':
                continue
            if len(tags) > 1:
                print("len(%s) == %d" % (key, len(tags)))
            if key == 'action_details':
                for a in tags:
                    if 'onclick' in a.attrs:
                        item[key].append(onclick_link(a['onclick']))
                continue
            if key == 'attachments' and len(tags):
                chunk = tags[0]
                attachments = []
                for anchor in chunk.find_all('a'):
                    name = anchor.text.strip()
                    link = anchor['href']
                    attachments.append((name, link))
                item[key] = attachments
                continue
            if not len(tags) and key == 'attachments':
                continue
            value = tags[0].text.strip()
            if value:
                item[key] = value
            else:
                item[key] = None
        return item

    def collect(self):
        self.retrieve_page(self.url)
        if b'Invalid parameters!' in self.content:
            item = dict()
            item['action_details'] = []
            item['bad_url'] = self.url
            self.item = item
            self.result = self.item
            print("Invalid parameters found", self.result)
            return
        self.item = self._get_item(self.soup)
        for key in ['passed', 'introduced', 'on_agenda']:
            if key in self.item and not self.item[key]:
                self.item[key] = None
            else:
                self.item[key] = make_true_date(self.item[key])
        if len(self.item['action_details']):
            self.item['acted_on'] = True
        else:
            self.item['acted_on'] = False
        self.item['id'], self.item['guid'] = legistar_id_guid(self.url)
        self.result = self.item




class BaseManager(object):
    def __init__(self, session):
        self.session = session

    def set_session(self, session):
        self.session = session

    def _range_filter(self, query, column, start, end):
        query = query.filter(column >= start)
        query = query.filter(column <= end)
        return query

    def query(self):
        raise NotImplementedError("Override me")

    def get(self, id):
        return self.query().get(id)

    def all(self):
        return self.query().all()



import transaction

from ..database import Department, Person
from .base import BaseManager


class DepartmentManager(BaseManager):
    def query(self):
        return self.session.query(Department)

    def add(self, id, guid, name):
        with transaction.manager:
            dept = Department(id, guid)
            dept.name = name
            self.session.add(dept)
        return self.session.merge(dept)


class PersonManager(BaseManager):
    def query(self):
        return self.session.query(Person)

    def _notrans_add(self, data):
        p = Person()
        for key, value in list(data.items()):
            setattr(p, key, value)
        self.session.add(p)
        return p

    def add(self, data):
        with transaction.manager:
            p = self._notrans_add(data)
        return self.session.merge(p)

    def add_people(self, people):
        plist = []
        with transaction.manager:
            for pinfo in people:
                p = self._notrans_add(pinfo)
                plist.append(p)
        return [self.session.merge(p) for p in plist]



import re

from ..util import legistar_id_guid

from .base import BaseCollector


DATA_IDENTIFIERS = dict(
    firstname='_lblFirst',
    lastname='_lblLast',
    website='_hypWebSite',
    notes='_lblNotes',
    photo_link='_imgPhoto',
    )
# photo_link no longer on people page
#    photo_link='_imgPhoto'
#    )


# dave ware's page
# https://hattiesburg.legistar.com/PersonDetail.aspx?ID=83530&GUID=62048E45-5276-44A8-8413-19B46071AA96&Search=
# I will probably need to start using selenium to collect
# the information from legistar, as a full list of people
# is only available through a javascript post.
# dware_link = 'PersonDetail.aspx?ID=83530&GUID=62048E45-5276-44A8-8413-19B46071AA96&Search=' # noqa

class PeopleCollector(BaseCollector):
    def __init__(self):
        BaseCollector.__init__(self)
        self.people_url = 'https://hattiesburg.legistar.com/People.aspx'
        self.url_prefix = 'https://hattiesburg.legistar.com/'

    def _get_people_anchors(self):
        self.set_url(self.people_url)
        self.retrieve_page()
        marker = '_hypPerson'
        exp = re.compile('.+%s$' % marker)
        return self.soup.find_all('a', id=exp)

    def _get_people_links(self):
        anchors = self._get_people_anchors()
        people_links = []
        for anchor in anchors:
            people_links.append(anchor['href'])
        # FIXME: hacky way of putting former people
        # into database.  This is done manually until
        # a better way of accessing the website presents
        # itself.
        # people_links.append(dware_link)
        return people_links

    def get_people_ids(self):
        anchors = self._get_people_anchors()
        ids = {}
        for anchor in anchors:
            name = anchor.text.strip()
            id, guid = legistar_id_guid(anchor['href'])
            ids[id] = name
        return ids

    def _get_person_page(self, link):
        url = self.url_prefix + link
        self.set_url(url)
        self.retrieve_page()

    def _get_person(self, page):
        markers = DATA_IDENTIFIERS
        item_keys = list(markers.keys())
        item = {}.fromkeys(item_keys)
        for key in item_keys:
            if key == 'photo_link':
                no_pix = False
            exp = re.compile('.+%s$' % markers[key])
            tags = page.find_all('span', id=exp)
            ttype = 'span'
            if not tags:
                tags = page.find_all('a', id=exp)
                ttype = 'a'
            if not tags:
                tags = page.find_all('img', id=exp)
                ttype = 'img'
            if not tags:
                if ttype == 'img' and key == 'photo_link':
                    no_pix = True
                else:
                    raise RuntimeError("no tags found for %s" % key)
            if len(tags) > 1:
                print("len(%s) == %d" % (key, len(tags)))
            if key == 'photo_link':
                if not no_pix:
                    tag = tags[0]
                    item[key] = tag['src']
                else:
                    item[key] = None
                continue
            tag = tags[0]
            if key == 'website':
                # item[key] = tag['href']
                item[key] = tag.text.strip()
                continue
            item[key] = tag.text.strip()
            print(key, item[key])
        item['id'], item['guid'] = legistar_id_guid(self.url)
        return item

    def collect(self):
        # self.retrieve_page(self.url)
        # self.item = self._get_item(self.soup)
        people = []
        links = self._get_people_links()
        for link in links:
            self._get_person_page(link)
            person = self._get_person(self.soup)
            people.append(person)
        self.people = people
        self.result = self.people


if __name__ == "__main__":
    pc = PeopleCollector()
    # pc.collect()



import re

from ..util import legistar_id_guid

from .base import BaseCollector


class DeptCollector(BaseCollector):
    def __init__(self):
        BaseCollector.__init__(self)
        url = 'https://hattiesburg.legistar.com/Departments.aspx'
        self.set_url(url)

    def _get_depts(self, page):
        depts = []
        # each entry is a tuple (id, guid, name)
        anchors = page.find_all('a', id=re.compile('.+_hypBody$'))
        for anchor in anchors:
            id, guid = legistar_id_guid(anchor['href'])
            name = anchor.text.strip()
            depts.append((id, guid, name))
        return depts

    def collect(self):
        self.retrieve_page(self.url)
        self.depts = self._get_depts(self.soup)
        self.result = self.depts
